/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2008, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Willow Garage nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *  \author Sachin Chitta, Ioan Sucan
 *********************************************************************/

#include <ros/ros.h>
#include <planning_environment_msgs/GetRobotState.h>
#include <planning_environment_msgs/GetJointTrajectoryValidity.h>
#include <gtest/gtest.h>


TEST(EnvironmentServer, AllowableContactsTest)
{

  ROS_INFO("Running test");
  ros::NodeHandle rh;

  ros::service::waitForService("/environment_server_right_arm/get_trajectory_validity");
  planning_environment_msgs::GetJointTrajectoryValidity::Request req;
  planning_environment_msgs::GetJointTrajectoryValidity::Response res;
  ros::ServiceClient check_trajectory_validity_client_ = rh.serviceClient<planning_environment_msgs::GetJointTrajectoryValidity>("/environment_server_right_arm/get_trajectory_validity");

  //  ros::service::waitForService("/environment_server_right_arm/get_robot_state");
  ros::ServiceClient get_state_client_ = rh.serviceClient<planning_environment_msgs::GetRobotState>("/environment_server_right_arm/get_robot_state");      
  planning_environment_msgs::GetRobotState::Request request;
  planning_environment_msgs::GetRobotState::Response response;
  if(get_state_client_.call(request,response))
  {
    req.robot_state = response.robot_state;
  }
  else
  {
    ROS_ERROR("Service call to get robot state failed on %s",get_state_client_.getService().c_str());
  }

  req.trajectory.joint_names.push_back("r_shoulder_pan_joint");
  req.trajectory.points.resize(100);
  for(unsigned int i=0; i < 100; i++)
  {    
        req.trajectory.points[i].positions.push_back((M_PI*i/15.0)/100.0);
    //req.trajectory.points[i].positions.push_back(0.0);
  }
  req.check_collisions = true;
  EXPECT_TRUE(check_trajectory_validity_client_.call(req,res));
  if(!(res.error_code.val == res.error_code.SUCCESS))
  {
    ROS_INFO("Collision detected without allowable contact region set");
  }
  EXPECT_FALSE(res.error_code.val == res.error_code.SUCCESS);

  sleep(2.0);
  req.allowed_contacts.resize(1);
  req.allowed_contacts[0].name = "table_region";
  req.allowed_contacts[0].shape.type = geometric_shapes_msgs::Shape::BOX;
  req.allowed_contacts[0].shape.dimensions.resize(3);

  req.allowed_contacts[0].shape.dimensions[0]= 1.0;
  req.allowed_contacts[0].shape.dimensions[1]= 1.0;
  req.allowed_contacts[0].shape.dimensions[2]= 1.0;
  req.allowed_contacts[0].pose_stamped.header.frame_id = "torso_lift_link";
  req.allowed_contacts[0].pose_stamped.header.stamp = ros::Time(0.0);

  req.allowed_contacts[0].pose_stamped.pose.position.x = 1.0;
  req.allowed_contacts[0].pose_stamped.pose.position.y = 0.0;
  req.allowed_contacts[0].pose_stamped.pose.position.z = 0.0;

  req.allowed_contacts[0].pose_stamped.pose.orientation.w = 1.0;
  req.allowed_contacts[0].link_names.push_back("r_shoulder_pan_link");
  req.allowed_contacts[0].link_names.push_back("r_shoulder_lift_link");
  req.allowed_contacts[0].link_names.push_back("r_upper_arm_link");
  req.allowed_contacts[0].link_names.push_back("r_elbow_flex_link");
  req.allowed_contacts[0].link_names.push_back("r_forearm_link");
  req.allowed_contacts[0].link_names.push_back("r_forearm_roll_link");
  req.allowed_contacts[0].link_names.push_back("r_wrist_flex_link");
  req.allowed_contacts[0].link_names.push_back("r_wrist_roll_link");
  req.allowed_contacts[0].link_names.push_back("r_gripper_l_finger_link");
  req.allowed_contacts[0].link_names.push_back("r_gripper_palm_link");
  req.allowed_contacts[0].penetration_depth = 3.0;
  
  EXPECT_TRUE(check_trajectory_validity_client_.call(req,res));
  if(res.error_code.val == res.error_code.SUCCESS)
  {
    ROS_INFO("Collision not detected with allowable contact region set");
  }
  EXPECT_TRUE(res.error_code.val == res.error_code.SUCCESS);
}

int main(int argc, char **argv){
  testing::InitGoogleTest(&argc, argv);
  ros::init (argc, argv, "allowable_contacts_test");
  return RUN_ALL_TESTS();
}
