/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2008, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Willow Garage nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *  \author Sachin Chitta, Ioan Sucan
 *********************************************************************/

#include <ros/ros.h>
#include <tf/tf.h>
#include <gtest/gtest.h>
#include <boost/thread.hpp>
#include <kinematics_msgs/GetPositionFK.h>
#include <motion_planning_msgs/GetMotionPlan.h>
#include <planning_environment_msgs/GetRobotState.h>

#include <actionlib/client/simple_action_client.h>
#include <move_arm_msgs/MoveArmAction.h>
#include <geometric_shapes_msgs/Shape.h>

void spinThread()
{
  ros::spin();
}

TEST(OMPL, PathConstraintsTest)
{
  ros::NodeHandle nh;
  ros::NodeHandle private_handle("~");
  boost::thread spin_thread(&spinThread);

  actionlib::SimpleActionClient<move_arm_msgs::MoveArmAction> move_arm(nh, "move_right_arm");
  move_arm.waitForServer();
  ROS_INFO("Connected to server");


  move_arm_msgs::MoveArmGoal goalA;
  goalA.motion_plan_request.group_name = "right_arm";
  goalA.motion_plan_request.num_planning_attempts = 1;
  private_handle.param<std::string>("planner_id",goalA.motion_plan_request.planner_id,std::string(""));
  private_handle.param<std::string>("planner_service_name",goalA.planner_service_name,std::string("/ompl_planning/plan_kinematic_path"));
  goalA.motion_plan_request.allowed_planning_time = ros::Duration(5.0);
    
  goalA.motion_plan_request.goal_constraints.set_position_constraints_size(1);
  goalA.motion_plan_request.goal_constraints.position_constraints[0].header.stamp = ros::Time::now();
  goalA.motion_plan_request.goal_constraints.position_constraints[0].header.frame_id = "torso_lift_link";
    
  goalA.motion_plan_request.goal_constraints.position_constraints[0].link_name = "r_wrist_roll_link";
  goalA.motion_plan_request.goal_constraints.position_constraints[0].position.x = 0.75;
  goalA.motion_plan_request.goal_constraints.position_constraints[0].position.y = -0.188;
  goalA.motion_plan_request.goal_constraints.position_constraints[0].position.z = 0;
    
  goalA.motion_plan_request.goal_constraints.position_constraints[0].constraint_region_shape.type = geometric_shapes_msgs::Shape::BOX;
  goalA.motion_plan_request.goal_constraints.position_constraints[0].constraint_region_shape.dimensions.push_back(0.02);
  goalA.motion_plan_request.goal_constraints.position_constraints[0].constraint_region_shape.dimensions.push_back(0.02);
  goalA.motion_plan_request.goal_constraints.position_constraints[0].constraint_region_shape.dimensions.push_back(0.02);

  goalA.motion_plan_request.goal_constraints.position_constraints[0].constraint_region_orientation.w = 1.0;
  goalA.motion_plan_request.goal_constraints.position_constraints[0].weight = 1.0;

  goalA.motion_plan_request.goal_constraints.set_orientation_constraints_size(1);
  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].header.stamp = ros::Time::now();
  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].header.frame_id = "torso_lift_link";    
  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].link_name = "r_wrist_roll_link";
  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].orientation.x = 0.0;
  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].orientation.y = 0.0;
  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].orientation.z = 0.0;
  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].orientation.w = 1.0;
    
  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].absolute_roll_tolerance = 0.04;
  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].absolute_pitch_tolerance = 0.04;
  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].absolute_yaw_tolerance = 0.04;

  goalA.motion_plan_request.goal_constraints.orientation_constraints[0].weight = 1.0;

  int num_test_attempts = 0;
  int max_attempts = 5;
  bool success = false;
  while (nh.ok())
  {
    bool finished_within_time = false;
    move_arm.sendGoal(goalA);
    finished_within_time = move_arm.waitForResult(ros::Duration(100.0));
    actionlib::SimpleClientGoalState state = move_arm.getState();
    success = (state == actionlib::SimpleClientGoalState::SUCCEEDED);
    if ((!finished_within_time || !success) && num_test_attempts < max_attempts)
    {
      move_arm.cancelGoal();
      ROS_INFO("Timed out achieving goal A, trying again. Trying again, attempt: %d",num_test_attempts);
      num_test_attempts++;
    }
    else
    {
      if(!success)
      {
        ROS_INFO("Action finished: %s",state.toString().c_str());
        move_arm.cancelGoal();
      }
      ROS_INFO("Action finished: %s",state.toString().c_str());
      break;
    }
  }



  ROS_INFO("Connected to server");
  ros::service::waitForService("ompl_planning/plan_kinematic_path");
  ros::service::waitForService("pr2_right_arm_kinematics/get_fk");
  ros::service::waitForService("environment_server_right_arm/get_robot_state");

  ros::ServiceClient get_motion_plan_client_ = nh.serviceClient<motion_planning_msgs::GetMotionPlan>("ompl_planning/plan_kinematic_path");
  
  motion_planning_msgs::GetMotionPlan::Request  request;
  motion_planning_msgs::GetMotionPlan::Response response;

  ros::ServiceClient get_state_client_ = nh.serviceClient<planning_environment_msgs::GetRobotState>("environment_server_right_arm/get_robot_state");      
  planning_environment_msgs::GetRobotState::Request request_state;
  planning_environment_msgs::GetRobotState::Response response_state;
  if(get_state_client_.call(request_state,response_state))
  {
    request.motion_plan_request.start_state = response_state.robot_state;
    request.motion_plan_request.start_state.joint_state.header.frame_id = "base_footprint";
  }
  else
  {
    ROS_ERROR("Service call to get robot state failed on %s",get_state_client_.getService().c_str());
  }

  std::vector<std::string> names(7);
  names[0] = "r_shoulder_pan_joint";
  names[1] = "r_shoulder_lift_joint";
  names[2] = "r_upper_arm_roll_joint";
  names[3] = "r_elbow_flex_joint";
  names[4] = "r_forearm_roll_joint";
  names[5] = "r_wrist_flex_joint";
  names[6] = "r_wrist_roll_joint";

  request.motion_plan_request.group_name = "right_arm";
  request.motion_plan_request.num_planning_attempts = 1;
  request.motion_plan_request.allowed_planning_time = ros::Duration(5.0);
  request.motion_plan_request.workspace_parameters.workspace_region_pose.header.stamp = ros::Time::now();

  private_handle.param<std::string>("planner_id",request.motion_plan_request.planner_id,std::string(""));    
  request.motion_plan_request.goal_constraints.joint_constraints.resize(names.size());

  for (unsigned int i = 0 ; i < request.motion_plan_request.goal_constraints.joint_constraints.size(); ++i)
  {
    request.motion_plan_request.goal_constraints.joint_constraints[i].joint_name = names[i];
    request.motion_plan_request.goal_constraints.joint_constraints[i].position = 0.0;
    request.motion_plan_request.goal_constraints.joint_constraints[i].tolerance_below = 0.1;
    request.motion_plan_request.goal_constraints.joint_constraints[i].tolerance_above = 0.1;
  }
    
  request.motion_plan_request.goal_constraints.joint_constraints[0].position = -2.0;
  request.motion_plan_request.goal_constraints.joint_constraints[3].position = -0.2;
  request.motion_plan_request.goal_constraints.joint_constraints[5].position = -0.2;

  request.motion_plan_request.path_constraints.orientation_constraints.resize(1);
  request.motion_plan_request.path_constraints.orientation_constraints[0].header.frame_id = "torso_lift_link";
  request.motion_plan_request.path_constraints.orientation_constraints[0].header.stamp = ros::Time::now();
  request.motion_plan_request.path_constraints.orientation_constraints[0].link_name = "r_wrist_roll_link";
   
  request.motion_plan_request.path_constraints.orientation_constraints[0].orientation.x = -0.167;
  request.motion_plan_request.path_constraints.orientation_constraints[0].orientation.y = -0.107;
  request.motion_plan_request.path_constraints.orientation_constraints[0].orientation.z = -0.825;
  request.motion_plan_request.path_constraints.orientation_constraints[0].orientation.w = 0.530;

  request.motion_plan_request.path_constraints.orientation_constraints[0].type = motion_planning_msgs::OrientationConstraint::HEADER_FRAME;
  request.motion_plan_request.path_constraints.orientation_constraints[0].absolute_roll_tolerance = 0.5;
  request.motion_plan_request.path_constraints.orientation_constraints[0].absolute_pitch_tolerance = 0.5;
  request.motion_plan_request.path_constraints.orientation_constraints[0].absolute_yaw_tolerance = M_PI;


  EXPECT_TRUE(get_motion_plan_client_.call(request,response));
  EXPECT_TRUE(!response.trajectory.joint_trajectory.points.empty());
  ROS_INFO("Output path has %d points",(int) response.trajectory.joint_trajectory.points.size());
  ros::ServiceClient fk_client = nh.serviceClient<kinematics_msgs::GetPositionFK>("pr2_right_arm_kinematics/get_fk");

  kinematics_msgs::GetPositionFK::Request  fk_request;
  kinematics_msgs::GetPositionFK::Response fk_response;
  fk_request.header.frame_id = "base_link";
  fk_request.fk_link_names.resize(1);
  fk_request.fk_link_names[0] = "r_wrist_roll_link";
  fk_request.robot_state.joint_state.position.resize(names.size());
  fk_request.robot_state.joint_state.name = names;

  btQuaternion constraint_orientation_quaternion;
  tf::quaternionMsgToTF(request.motion_plan_request.path_constraints.orientation_constraints[0].orientation,constraint_orientation_quaternion);
  btMatrix3x3 constraint_orientation = btMatrix3x3(constraint_orientation_quaternion);
  

  // Now, check the returned trajectory to make sure it satisfies the constraints
  for(unsigned int i=0; i < response.trajectory.joint_trajectory.points.size(); i++)
  {
    fk_request.robot_state.joint_state.position = response.trajectory.joint_trajectory.points[i].positions;
    EXPECT_TRUE(fk_client.call(fk_request, fk_response));

    btQuaternion actual_orientation_quaternion;
    tf::quaternionMsgToTF(fk_response.pose_stamped[0].pose.orientation,actual_orientation_quaternion);
    btMatrix3x3 actual_orientation = btMatrix3x3(actual_orientation_quaternion);
    btMatrix3x3 result = actual_orientation * constraint_orientation.inverse();

    double roll, pitch, yaw;
    result.getRPY(roll, pitch, yaw);

    EXPECT_TRUE(abs(roll) < request.motion_plan_request.path_constraints.orientation_constraints[0].absolute_roll_tolerance);
    EXPECT_TRUE(abs(pitch) < request.motion_plan_request.path_constraints.orientation_constraints[0].absolute_pitch_tolerance);
    EXPECT_TRUE(abs(yaw) < request.motion_plan_request.path_constraints.orientation_constraints[0].absolute_yaw_tolerance);

  }
  ros::shutdown();
  spin_thread.join();
}

int main(int argc, char **argv){
  testing::InitGoogleTest(&argc, argv);
  ros::init(argc,argv,"path_constraints_test");
  return RUN_ALL_TESTS();
}
