//Software License Agreement (BSD License)

//Copyright (c) 2008, Willow Garage, Inc.
//All rights reserved.

//Redistribution and use in source and binary forms, with or without
//modification, are permitted provided that the following conditions
//are met:

// * Redistributions of source code must retain the above copyright
//   notice, this list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above
//   copyright notice, this list of conditions and the following
//   disclaimer in the documentation and/or other materials provided
//   with the distribution.
// * Neither the name of Willow Garage, Inc. nor the names of its
//   contributors may be used to endorse or promote products derived
//   from this software without specific prior written permission.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
//COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
//BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
//CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
//LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
//ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
//POSSIBILITY OF SUCH DAMAGE.

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ros/ros.h>
#include <kinematics_msgs/GetPositionIK.h>
#include <kinematics_msgs/GetPositionFK.h>
#include <kinematics_msgs/GetKinematicTreeInfo.h>
#include <kinematics_msgs/GetCollisionFreePositionIK.h>

#include <gtest/gtest.h>

#define IK_NEAR 1e-4
#define IK_NEAR_TRANSLATE 1e-5

static const std::string ARM_FK_NAME = "/pr2_ik_right_arm/get_fk";
static const std::string ARM_IK_NAME = "/pr2_ik_right_arm/get_ik";
static const std::string ARM_COLLISION_IK_NAME = "/pr2_ik_right_arm/get_collision_free_ik";
static const std::string ARM_QUERY_NAME = "/pr2_ik_right_arm/get_kinematic_tree_info";

static const int NUM_TESTS = 1000;

double gen_rand(double min, double max)
{
  int rand_num = rand()%100+1;
  double result = min + (double)((max-min)*rand_num)/101.0;
  return result;
}

bool NOT_NEAR(const double &v1, const double &v2, const double &NEAR)
{
  if(fabs(v1-v2) > NEAR)
    return true;
  return false;
}

TEST(PR2ArmIKWithCollisionNode, inverseKinematics)
{
  srand ( time(NULL) ); // initialize random seed
  ros::NodeHandle rh;
  ros::service::waitForService(ARM_QUERY_NAME);
  ros::service::waitForService(ARM_FK_NAME);
  ros::service::waitForService(ARM_COLLISION_IK_NAME);

  ros::Duration(20.0).sleep();

  ros::ServiceClient query_client = rh.serviceClient<kinematics_msgs::GetKinematicTreeInfo>(ARM_QUERY_NAME);
  ros::ServiceClient fk_client = rh.serviceClient<kinematics_msgs::GetPositionFK>(ARM_FK_NAME);
  ros::ServiceClient ik_with_collision_client = rh.serviceClient<kinematics_msgs::GetCollisionFreePositionIK>(ARM_COLLISION_IK_NAME);
  ros::ServiceClient ik_client = rh.serviceClient<kinematics_msgs::GetPositionIK>(ARM_IK_NAME);

  // define the service messages
  kinematics_msgs::GetKinematicTreeInfo::Request request;
  kinematics_msgs::GetKinematicTreeInfo::Response response;
	    
  bool service_call = query_client.call(request, response);
  ASSERT_TRUE(service_call);
  ASSERT_TRUE(response.kinematic_tree_info.joint_names.size() == response.kinematic_tree_info.limits.size());
  ASSERT_TRUE(!response.kinematic_tree_info.link_names.empty());

  int num_joints;
  std::vector<double> min_limits, max_limits;
  num_joints = (int) response.kinematic_tree_info.joint_names.size();
  for(int i=0; i< num_joints; i++)
  {
    min_limits.push_back(response.kinematic_tree_info.limits[i].min_position);
    max_limits.push_back(response.kinematic_tree_info.limits[i].max_position);
  }

  kinematics_msgs::GetCollisionFreePositionIK::Request  ik_request;
  kinematics_msgs::GetCollisionFreePositionIK::Response ik_response;

  ik_request.ik_request.ik_seed_state.joint_state.name = response.kinematic_tree_info.joint_names;
  ik_request.ik_request.ik_seed_state.joint_state.position.resize(num_joints);
  ik_request.ik_request.ik_link_name = response.kinematic_tree_info.link_names[0];
  ik_request.timeout = ros::Duration(2.0);
  for(int i=0; i< num_joints; i++)
  {
    ik_request.ik_request.ik_seed_state.joint_state.position[i] = gen_rand(std::max(min_limits[i],-M_PI),std::min(max_limits[i],M_PI));
  }

  ik_request.ik_request.ik_link_name = "r_wrist_roll_link";
  ik_request.ik_request.pose_stamped.header.frame_id = "torso_lift_link";
  ik_request.ik_request.pose_stamped.pose.position.x = 0.45;
  ik_request.ik_request.pose_stamped.pose.position.y = 0.0;
  ik_request.ik_request.pose_stamped.pose.position.z = 0.0;

  ik_request.ik_request.pose_stamped.pose.orientation.x = 0.0;
  ik_request.ik_request.pose_stamped.pose.orientation.y = 0.0;
  ik_request.ik_request.pose_stamped.pose.orientation.z = 0.0;
  ik_request.ik_request.pose_stamped.pose.orientation.w = 1.0;

  bool ik_service_call = ik_with_collision_client.call(ik_request,ik_response);
  if(!ik_service_call)
  {
    ROS_ERROR("IK service call failed");
  }
  ROS_INFO("IK response code: %d",(int)ik_response.error_code.val);
  ROS_DEBUG("IK response: num positions %d, num solution joints %d",(int)ik_response.solution.joint_state.position.size(),(int)ik_response.solution.joint_state.name.size());
  ASSERT_TRUE(ik_response.error_code.val == ik_response.error_code.IK_LINK_IN_COLLISION);


  kinematics_msgs::GetPositionIK::Request  pos_ik_request;
  kinematics_msgs::GetPositionIK::Response pos_ik_response;
  pos_ik_request.ik_request = ik_request.ik_request;
  pos_ik_request.timeout = ik_request.timeout;
  bool ik_call = ik_client.call(pos_ik_request,pos_ik_response);
  ASSERT_TRUE(ik_call);
  ASSERT_TRUE(pos_ik_response.error_code.val == pos_ik_response.error_code.SUCCESS);
  ros::shutdown();
}

int main(int argc, char **argv){
  testing::InitGoogleTest(&argc, argv);
  ros::init (argc, argv, "pr2_ik_with_collision_regression_test");
  return RUN_ALL_TESTS();
}
